from .perftester import (
    IncorrectArgumentError,
    TimeTestError,
    MemoryTestError,
    FunctionError,
    CLIPathError,
    LogFilePathError,
    Config,
    config,
    memory_usage_benchmark,
    memory_usage_test,
    time_benchmark,
    time_test,
    pp,
)



import ast
import copy
import inspect
import json
import pathlib
import pickle
import rounder

import easycheck

from collections import namedtuple
from copy import deepcopy
from functools import wraps
from memory_profiler import memory_usage
from time import perf_counter
from typing import Callable, Iterable, Union


# Globals, used as dict fields in calls

CALLEDFROM = "called from"
CALLMODE = "calls"
INSTANCEMODE = "instances"
CALLS = "no of calls"
INSTANCES = "no of instances"
MEM = "memory peak"
TIME = "execution time"
TIME_ALL = "time spent inside (all)"
TIME_MEAN = "time spent inside (mean)"


# Exceptions

class IncorrectDictRepresentationError(Exception):
    pass


# Performance decorators

def time_performance(func: Callable) -> Callable:
    """Time performance decorator.
    
    It is used to analyze the performance of the decorated function in terms
    of execution time.
    """
    @wraps(func)
    def inner(*args, **kwargs):
        _type, function = _get_type_and_name(func)
        if _type == "class method":
            aargs = list(args)
            aargs[0] = "self"
            aargs = tuple(aargs)
            arguments = _stringify_args_kwargs(aargs, kwargs)
        else:
            arguments = _stringify_args_kwargs(args, kwargs)
        
        global calls
        
        call_type = INSTANCEMODE if _type == "class" else CALLMODE
        call_count = INSTANCES if _type == "class" else CALLS
        
        if not calls.registered[_type].get(function, False):
            calls.registered[_type][function] = {}
            calls.registered[_type][function][call_type] = {}

        calls.registered[_type][function][call_count] = (
            calls.registered[_type][function].get(call_count, 0) + 1
        )
        
        start = perf_counter()
        ret = func(*args, **kwargs)
        end = perf_counter()

        ID = calls.registered[_type][function][call_count]
        calls.registered[_type][function][call_type][ID] = {
            CALLEDFROM: inspect.stack()[1][3],
        }
        calls.registered[_type][function][call_type][ID][TIME] = (
            rounder.signif_object(end - start, 5)
        )
        calls.registered[_type][function][call_type][ID]["call"] = f"{function}({arguments})"
        calls.registered[_type][function][TIME_ALL] = (
            calls
                .registered[_type][function]
                .get(TIME_ALL, 0)
            + (end - start)
        )
        calls.registered[_type][function][TIME_MEAN] = (
            calls.registered[_type][function][TIME_ALL] /
            (
                calls.registered[_type][function].get(CALLS, None)
                or calls.registered[_type][function][INSTANCES]
            )
        )
        return ret

    return inner


def memory_performance(func: Callable) -> Callable:
    """Memory performance decorator.
    
    It is used to analyze the performance of the decorated function in terms
    of memory usage.
    """

    def inner(*args, **kwargs):
        _type, function = _get_type_and_name(func)
        if _type == "class method":
            aargs = list(args)
            aargs[0] = "self"
            aargs = tuple(aargs)
            arguments = _stringify_args_kwargs(aargs, kwargs)
        else:
            arguments = _stringify_args_kwargs(args, kwargs)
        
        global calls
        
        call_type = INSTANCEMODE if _type == "class" else CALLMODE
        call_count = INSTANCES if _type == "class" else CALLS
        
        if not calls.registered[_type].get(function, False):
            calls.registered[_type][function] = {}
            calls.registered[_type][function][call_type] = {}

        calls.registered[_type][function][call_count] = (
            calls.registered[_type][function].get(call_count, 0) + 1
        )
        memory_results, ret = memory_usage((func, args, kwargs), retval=True)
        peak_memory = min(memory_results)
        
        ID = calls.registered[_type][function][call_count]
        calls.registered[_type][function][call_type][ID] = {
            CALLEDFROM: inspect.stack()[1][3],
        }
        calls.registered[_type][function][call_type][ID][MEM] = (
            peak_memory
        )
        calls.registered[_type][function][call_type][ID]["call"] = f"{function}({arguments})"
        calls.registered[_type][function][MEM] = max(
            # current max
            calls
                .registered[_type][function]
                .get(MEM, 0),
            # current memory_peak
            peak_memory
        )
        return ret

    return inner


# Class containing all the registered calls and class instances.

class Calls:
    __slots__ = ("registered", "analyze", "save", "read")
    
    def __init__(self):
        self.registered: dict = {"class": {}, "class method": {}, "function": {}}
    
    def show(self, digits=4, indent=4):
        rounded_calls = rounder.signif_object(self.registered,
                                              use_copy=True,
                                              digits=digits)
        print("Classes:\n", json.dumps(rounded_calls["class"], indent=indent))
        print("Class methods:\n", json.dumps(rounded_calls["class method"], indent=indent))
        print("Functions:\n", json.dumps(rounded_calls["function"], indent=indent))

    def __repr__(self):
        if not self.registered.keys():
            return "No registered calls"
        
        instances = self._get_count("class")
        method_calls = self._get_count("class method")
        func_calls = self._get_count("function")
        
        return (
            "Registered:\n"
            f"  * {instances} class instances\n"
            f"  * {method_calls} class methods\n"
            f"  * {func_calls} function calls\n"
        )
    
    def _get_count(self, which):
        calls = 0
        if self.registered[which]:
            for k in self.registered[which]:
                key = INSTANCES if which == "class" else CALLS
                calls += self.registered[which][k][key]
        return calls


class CallsAnalyzer:
    """Class to analyze Calls.registered.
    
    This is a simple analyzer. A more adanced one will be offered
    in a dedicated understand extension.
    """
    def _update_calls(self):
        global calls
        self.calls = calls.registered

    def summarize(self, digits=4) -> dict:
        self._update_calls()
        summary_calls = copy.deepcopy(self.calls)
        for key in ("class method", "function"):
            if summary_calls[key]:
                for k in summary_calls[key]:
                    _ = summary_calls[key][k].pop(CALLMODE)
        if summary_calls["class"]:
            for k in summary_calls["class"]:
                _ = summary_calls["class"][k].pop(INSTANCEMODE)
            
        summary_calls = rounder.signif_object(summary_calls, digits=digits)
        return summary_calls

    summarise = summarize
    

class CallsSaver:
    def _update(self):
        global calls
        self.calls = calls.registered

    def to_json(self, path: Union[str, pathlib.Path]):
        self._update()
        json_dict = json.dumps(self.calls)
        with open(path, "w") as json_file:
            json_file.write(json_dict)
    
    def to_text(self, path: Union[str, pathlib.Path]):
        self._update()
        with open(path, "w") as text_file:
            text_file.write(str(self.calls))

    def to_pickle(self, path: Union[str, pathlib.Path]):
        self._update()
        with open(path, "wb") as pickle_file:
            pickle.dump(self.calls, pickle_file, protocol=pickle.HIGHEST_PROTOCOL)


class IncorrectCallsInstanceError(Exception): ...


CallsInstance = namedtuple("CallsInstance", "type registered")


class CallsReader:
    __slots__ = tuple()

    def _check_calls_object(self, obj, path):
        # check if obj is a valid Calls object
        easycheck.check_type(
            obj,
            expected_type=dict,
            handle_with=IncorrectDictRepresentationError,
            message=f"Object parsed from {path} is not a dict but"
                    f" a {type(obj).__name__}"
        )
        # check if the required keys are there
        return self._get_type_of_calls(obj)

    @staticmethod
    def _get_type_of_calls(obj):
        # check if this is a time or memory calls dict
        # if time: return "time"
        # if memory: return "memory"
        # else: raise IncorrectCallsInstanceError
        for _type in ["class", "class method", "function"]:
            if obj[_type].keys():
                first = list(obj[_type].keys())[0]
                if TIME_ALL in obj[_type][first].keys():
                    return "time"
                if MEM in obj[_type][first].keys():
                    return "memory"
        raise IncorrectCallsInstanceError(
            "The object is neither Time nor Memory Calls dictionary."
            )

    def from_json(self, path: Union[str, pathlib.Path]):
        with open(path) as jsonfile:
            registered = json.load(jsonfile)
        
        # use int IDs (used as keys) instead of str,
        # which is used by json.load
        for _type in ["class", "class method", "function"]:
            calls = "instances" if _type == "class" else "calls"
            for method in registered[_type]:
                registered[_type][method][calls] = {
                    int(instance): value
                    for instance, value in registered[_type][method][calls].items()
                }
        return CallsInstance(
            type=self._check_calls_object(registered, path),
            registered=registered
        )

    def from_text(self, path: Union[str, pathlib.Path]):
        with open(path) as f:
            try:
                obj_from_text = ast.literal_eval(f.read())
            except Exception as e:
                raise IncorrectDictRepresentationError from e
        self._check_calls_object(obj_from_text, path)
        return CallsInstance(
            type=self._check_calls_object(obj_from_text, path),
            registered=obj_from_text
        )

    def from_pickle(self, path: Union[str, pathlib.Path]):
        with open(path, "rb") as pickle_file:
            pkl = pickle.load(pickle_file)
        return CallsInstance(
            type=self._check_calls_object(pkl, path),
            registered=pkl
        )


# Helpers

def _is_class(obj):
    return repr(obj).startswith("<class")


def _is_this_function_a_method(func):
    return "." in func.__qualname__


def _get_type_and_name(obj):
    if _is_class(obj):
        return "class", f"{obj.__module__}.{obj.__name__}"
    if hasattr(obj, "__wrapped__"):
        name = obj.__wrapped__.__name__
    if _is_this_function_a_method(obj):
        name = obj.__qualname__
        return "class method", f"{obj.__module__}.{name}"
    return "function", f"{obj.__module__}.{obj.__name__}"


def type_and_len_of_iterable(it, max_len=1000):
    """Get length of an iterable.
    
    The iterable is copied first, and the function works
    on the deep copy. For generators expressions, the function
    returns "generator" as cannot get its length without emptying
    the original generator.
    
    If maximum_len is reached, instead of the actual length, the function
    returns "over {max_len}", to make the function cheap.
    """
    # Get length, based on copy
    it_copy = deepcopy(it)
    
    c = 0
    max_len_reached = False
    for i in it_copy:
        c += 1
        if i == max_len:
            max_len_reached = True
            break

    if max_len_reached:
        return f"{type(it).__name__} of over {max_len} items"
    if not c:
        return f"empty {type(it).__name__}"
    return f"{type(it).__name__} of {c} item{'s' if c > 1 else ''}"


def _add_quotes_to_str(obj):
    """Add quotes to strings.

    The function works recursively, so that it goes deep into the structure
    to reach any string. If obj is "self", the quotes are not added, so that
    in case of class methods self is printed without quotes.

    Warning: Note that in a dictionary, it will add quotes only to values,
    not to keys.

    >>> _add_quotes_to_str(222)
    222
    >>> _add_quotes_to_str("222")
    "'222'"
    >>> _add_quotes_to_str(("222", ))
    'tuple of 1 item'
    >>> _add_quotes_to_str(("222", 50, "String", 'Also string'))
    'tuple of 4 items'
    >>> _add_quotes_to_str(("222", 50, ["String", 'Also string']))
    'tuple of 3 items'
    >>> _add_quotes_to_str((["'whatever'", "'and whatever'"], 20))
    'tuple of 2 items'
    >>> _add_quotes_to_str({"a": "whatever", "b":"Something"})
    {'a': "'whatever'", 'b': "'Something'"}
    """
    if obj == "self":
        return f"self"
    if not obj:
        return obj

    if isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = _add_quotes_to_str(v)
        return obj
    if isinstance(obj, str):
        return f"'{obj}'"
    if isinstance(obj, range):
        n = len(obj)
        return f"range of {n} item{'s' if n > 1 else ''}"
    if isinstance(obj, Iterable):
        # For generator expression, we need to use itertools.tee,
        # but this cannot be done behind the scenes, as the original
        # object is empty and cannot be thus used.
        try:
            return type_and_len_of_iterable(obj)
        except TypeError:
            # cannot pickle the generator type,
            # hence we do not get its length
            return "generator"

    return obj


def _stringify_args_kwargs(args, kwargs, digits=4):
    """Make a string out of args and kwargs.

    The final string is formatted in a way that can be used in a print
    of a call to a function. So, for instance, white spaces are added after
    each argument; quotes are added to strings, etc.

    >>> _stringify_args_kwargs(("something", 20), {})
    "'something', 20"
    >>> _stringify_args_kwargs(tuple(), {"x": 300})
    'x=300'
    >>> _stringify_args_kwargs(tuple(), {"x": "mixed"})
    "x='mixed'"
    >>> _stringify_args_kwargs(("something", 20), {"x": 300})
    "'something', 20, x=300"
    >>> _stringify_args_kwargs(("something", 20), {"x": "something else"})
    "'something', 20, x='something else'"
    """
    arguments = ""
    if len(args) > 0 and len(kwargs) == 0:
        if len(args) == 1:
            a = _add_quotes_to_str(rounder.signif_object(args[0], digits))
            arguments = f"{a}"
        else:
            to_join = [str(_add_quotes_to_str(rounder.signif_object(a, digits))) for a in args]
            arguments = f"{', '.join(to_join)}"
    elif len(kwargs) > 0 and len(args) == 0:
        a = ", ".join(f"{k}={str(_add_quotes_to_str(rounder.signif_object(v, digits)))}" for k, v in kwargs.items())
        arguments = f'{a}'
    elif len(kwargs) > 0 and len(args) > 0:
        if len(args) == 1:
            arguments = f"{_add_quotes_to_str(rounder.signif_object(args[0], digits))}"
        else:
            to_join = [str(_add_quotes_to_str(rounder.signif_object(a, digits))) for a in args]
            arguments = f"{', '.join(to_join)}"
        if len(kwargs) == 1:
            key = next(iter(kwargs))
            arguments += f", {key}={_add_quotes_to_str(rounder.signif_object(kwargs[key], digits))}"
        else:
            arguments += f', {", ".join(f"{k}={_add_quotes_to_str(rounder.signif_object(v, digits))}" for k, v in kwargs.items())}'
    return arguments


# The final creation of the instance of Calls

calls = Calls()
calls.analyze = CallsAnalyzer()
calls.save = CallsSaver()
calls.read = CallsReader()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from pprint import pprint


class Container:
    def __init__(self, values):
        self.values = values

    def __gt__(self, other):
        return Container(type(self.values)(other(i) for i in self.values))


class Show:
    def __call__(self, obj=None, func=None, *args, **kwargs):
        self.obj = obj
        if not func:
            pprint(obj.result)    
        else:
            pprint(func(obj.result, *args, **kwargs))
        return obj
    
    def __gt__(self, other):
        return other(self.obj.result)



class Value:
    def __init__(self, value):
        self.result = value

    def __call__(self):
        return self.result
    
    def __gt__(self, other):
        return other(self.result)


class Do:
    def __init__(self, func, *args, parallel=False, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.result = None
        self.parallel = parallel
    
    def __call__(self, *add_args, **add_kwargs):
        these_args = {**self.kwargs, **add_kwargs}
        self.result = self.func(*add_args, *self.args, **these_args)
        return self.result
    
    def __gt__(self, other):
        if isinstance(other, Show):
            return other(self)
        try:
            return other(self.__call__())
        except:
            return other(self.result)
    


double = lambda x: 2 * x
square = lambda x: x ** 2
add = lambda x, y: x + y

Value(10) > Do(double) > Show() > Do(double) > Show()

Value(10) > Do(double) > Do(square) > Do(add, 50) > Do(print)
Value(10) > Do(double) > Do(square) > Show() > Do(add, 50) > Show()

Container([1, 2, 3]) > Do(double) > Show()


class Value:
    def __init__(self, value):
        self.result = value

    def __call__(self):
        return self.result
    
    def __gt__(self, other):
        return other(self.result)


class Do:
    def __init__(self, func=None, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
    
    def __call__(self, *add_args, **add_kwargs):
        these_args = {**self.kwargs, **add_kwargs}
        self.result = self.func(*add_args, *self.args, **these_args)
        return self.result
    
    def __gt__(self, other):
        try:
            return other(self.__call__())
        except:
            return other(self.result)


double = lambda x: 2 * x
square = lambda x: x ** 2
add = lambda x, y: x + y

Value(10) > Do(double) > Do(square) > Do(add, 50) > Do(print)



"""Module responsible for the CLI perftester command."""
import importlib
import inspect
import os
import pathlib
import sys
from collections import namedtuple
from easycheck import check_if_paths_exist
from perftester import config, TimeTestError, MemoryTestError, CLIPathError


TestResults = namedtuple("TestResults", "passed_tests failed_tests")

config.full_traceback()


def main():
    files = _read_cli_args()
    _initialize_log_file(len(files))
    _import_settings_from_config_file()
    passed_perftesters, failed_perftesters = [], []

    for file in files:
        module, module_name = _import_module(file)
        perftester_functions = _find_perftester_functions(module)
        for func in perftester_functions:
            this_test = _perftester(module_name, func)
            if this_test:
                failed_perftesters.append(f"{str(module_name)}.{func.__name__}")
            else:
                passed_perftesters.append(f"{str(module_name)}.{func.__name__}")

    test_results = TestResults(
        passed_tests=passed_perftesters, failed_tests=failed_perftesters
    )
    _log_perftester_results(test_results)


def _read_cli_args():
    if len(sys.argv) == 1:
        path = pathlib.Path(os.getcwd())
    else:
        # sys.argv[1] is always a string, so no need to check it
        path = pathlib.Path(sys.argv[1])
    check_if_paths_exist(
        path, CLIPathError, "Incorrent path provided with perftester CLI command"
    )
    if path.is_dir():
        files = _find_perftester_files(path)
    elif path.is_file:
        files = [
            path,
        ]
    else:
        raise CLIPathError(
            f"Unexpected problem with path {path}, please double check"
        )
    return files


def _log(message):
    print(message)
    if config.log_to_file:
        try:
            with open(config.log_file, "a") as f:
                f.write(message + "\n")
                f.flush()
        except OSError as e:
            print(
                f"Error in writing to {config.log_file}: {e}. "
                "perftester log will not be saved there."
            )
            config.log_to_file = False


def _import_settings_from_config_file():
    settings_file = config.config_file
    if settings_file.exists():
        sys.path.append(str(settings_file.parent.absolute()))
        importlib.import_module("config_perftester")
        _log(f"Importing settings from {settings_file}.")
    else:
        _log(
            "No settings file detected, using default perftester configuration."
        )


def _initialize_log_file(files_len):
    if config.log_to_file:
        try:
            with open(config.log_file, "w") as f:
                f.write("")
        except OSError as e:
            print(
                f"Error in writing to {config.log_file}: {e}. "
                "perftester log will not be saved there."
            )
            config.log_to_file = False

    _log(
        "Performance tests using perftester\n"
        "perftester: https://github.com/nyggus/perftester\n"
        "--------------------------------------------"
        f"\n\nCollected {files_len} "
        f"perftester module{'s' if files_len > 1 else ''} for testing.\n"
    )


def _import_module(file):
    path_str = str(file.parent.absolute())
    if path_str not in sys.path:
        sys.path.append(path_str)
    module_name = file.name[:-3]
    module = importlib.import_module(module_name)
    return module, module_name


def _perftester(module_name, func, *args, **kwargs):
    try:
        func(*args, **kwargs)
    except TimeTestError as e:
        _log(f"\nTimeTestError in {module_name}.{func.__name__}\n{e}")
        return 1
    except MemoryTestError as e:
        _log(f"\nMemoryTestError in {module_name}.{func.__name__}\n{e}")
        return 1
    except Exception as e:
        _log(
            "\nUnexpected error in test function "
            f"{module_name}.{func.__name__}\n{e}\n"
        )
        return 1
    return 0


def _find_perftester_functions(module):
    functions = []
    items = dir(module)
    for item in items:
        item_real = getattr(module, item)
        if inspect.isfunction(item_real) and item.startswith("perftester_"):
            functions.append(item_real)
    return functions


def _find_perftester_files(path: pathlib.Path):
    return [file for file in path.rglob("perftester_*.py")]


def _log_perftester_results(test_results):
    passed_perftesters = test_results.passed_tests
    failed_perftesters = test_results.failed_tests
    n_of_tests = len(passed_perftesters) + len(failed_perftesters)

    _log(
        f"\n\nPerfomance testing done.\nOut of {n_of_tests} tests,"
        f" {len(passed_perftesters)} has passed"
        f" and {len(failed_perftesters)} has failed."
    )
    if len(passed_perftesters) > 0:
        _log("\nPassed tests:")
        for test in passed_perftesters:
            _log(test)
    if len(failed_perftesters) > 0:
        _log(f"\nFailed tests:")
        for test in failed_perftesters:
            _log(test)
    print()


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

extras_requirements = {
    "dev": ["wheel==0.37.1", "black"],
}

setuptools.setup(
    name="perftester",
    version="0.5.0",
    author="Nyggus",
    author_email="nyggus@gmail.com",
    description="Lightweight performance testing in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nyggus/perftester",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["easycheck", "rounder", "memory_profiler==0.60.0"],
    python_requires=">=3.8",
    extras_require=extras_requirements,
    entry_points={"console_scripts": ["perftester = perftester.__main__:main"]},
)



"""A lightweight module for simple performance testing in Python.

The singleton Config class is used in all testing in the same session; its
instance is created during import and can be used by all functions.

Do note that unlike timeit functions, perftester works with times of a single
call of a function. That way, the user can define tests without worrying of
different settings: while the "number" argument in timeit functions affects
the results, it does not so in perftester.

Remember that all lambda functions are stored as anonymous functions,
without a name. Hence it's better to avoid using lambdas in perftester.

For the sake of pretty-printing the benchmarks, perftester comes with a pp
function, which rounds all numbers to four significant digits and prints
the object using pprint.pprint:
>>> import perftester as pt
>>> pt.pp([2.1212, 2.93135])
[2.121, 2.931]

You can change this behavior, however:
>>> pt.config.digits_for_printing = 2
>>> pt.pp([2.1212, 2.93135])
[2.1, 2.9]

Let's return to previous settings:
>>> pt.config.digits_for_printing = 4
"""
import copy
import os
import rounder
import sys
import timeit

from collections import namedtuple
from collections.abc import Callable
from easycheck import (
    check_argument,
    check_if,
    check_if_not,
    check_instance,
    check_if_paths_exist,
    assert_instance,
)
from memory_profiler import memory_usage
from pathlib import Path
from pprint import pprint
from statistics import mean


class CLIPathError(Exception):
    """Exception class to be used for the CLI perftester app."""


class LogFilePathError(Exception):
    """Exception class to be used when incorrect path is provided for a log file."""


class LackingLimitsError(Exception):
    """Exception class to be used when no limits are set for a test."""


class IncorrectArgumentError(Exception):
    """Exception class to be used when function args are incorrect."""


class TimeTestError(Exception):
    """Exception class to be used when time_test() does not pass."""


class MemoryTestError(Exception):
    """Exception class to be used when memory_usage_test() does not pass."""


class FunctionError(Exception):
    """Exception class to be used when the tested code throws an error."""


# Configuration


class Config:
    """Default configuration for testing.

    It is a singleton whose instance is created once, during import
    of the perftester module.
    """
    config_file = Path("config_perftester.py")
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        """
        Attributes:
            defaults: a dict of dicts; defaults["time"] provide number and repeat
                to be passed to timeit.repeat; defaults["memory"] provide repeat,
                and the memory profiler will run that many times
            settings: a dict with values having the same structure as defaults;
                keys of the dict are functions; if the user does not change them,
                defaults are used
            time_benchmark: execution time of the in-built benchmark function
            memory_benchmark: memory usage of the in-built benchmark function
            digits_for_printing: how many digits should be used by rounder.signif()?
            log_to_file (bool): log results to a file or not?
            log_file (pathlib.Path): path to log file
        """
        self.defaults = {
            "time": {"number": 100_000, "repeat": 5},
            "memory": {"repeat": 1},
        }
        self.settings = {}
        self._benchmark_time()
        self._benchmark_memory()

        # Set up the number of digits to be used in rounder.signif_object(),
        # which rounds numbers to a significant number of digits
        self.digits_for_printing = 4

        self.log_to_file = True
        self.log_file = Path(os.getcwd()) / "perftester.log"

    @property
    def digits_for_printing(self):
        return self._digits_for_printing

    @digits_for_printing.setter
    def digits_for_printing(self, value):
        check_instance(
            value,
            int,
            message=f"Argument value must be an int, not {type(value).__name__}",
        )
        self._digits_for_printing = value

    @property
    def log_to_file(self):
        return self._log_to_file

    @log_to_file.setter
    def log_to_file(self, value):
        check_instance(
            value,
            bool,
            message=f"Argument value must be an int, not {type(value).__name__}",
        )
        self._log_to_file = value

    @property
    def log_file(self):
        return self._log_file

    @log_file.setter
    def log_file(self, path):
        path = Path(path)
        check_if_paths_exist(
            path.parent.absolute(),
            LogFilePathError,
            f"Argument path must be a valid path",
        )
        self._log_file = path

    @staticmethod
    def cut_traceback():
        """Remove traceback from exceptions, and report only exceptions.

        This is the default behavior of perftester, as when testing performance,
        you do not look for bugs in code, so no need to analyze traceback.
        """
        sys.tracebacklimit = 0

    @staticmethod
    def full_traceback():
        """Use full traceback to report exceptions."""
        sys.tracebacklimit = None

    def reload_time(self):
        """Reload built-in time benchmarks.

        This is done so that a new relative test is run against a built-in
        benchmark that has just been ran, not some time ago.
        """
        self._benchmark_time()

    def reload_memory(self):
        """Reload built-in memory benchmarks.

        This is done so that a new relative test is run against a built-in
        benchmark that has just been ran, not some time ago.

        WARNING: This method is NOT used in normal circumstances because memory
        usage checks do not change over time, so there is no need to update them.
        You need to use this method only when you change the benchmark function.
        """
        self._benchmark_memory()

    def get_setting(self, func, which, item):
        """Get setting (number or repeat) for a function.

        If they were not set up, the default value is provided.

        Args:
            func (Callable): a callable for which the setting is to be changed
            which (str): either "time" or "memory", for which the setting is
                to be changed
            item (str): either "repeat" or "number"
        """
        whiches = ("time", "memory")
        items = ("number", "repeat")
        check_instance(
            func,
            Callable,
            IncorrectArgumentError,
            message="Argument func must be a callable.",
        )
        check_argument(
            which,
            expected_type=str,
            expected_choices=whiches,
            handle_with=IncorrectArgumentError,
            message=(
                "Argument which must be str from among "
                f"{_str_iterable(whiches)}"
            ),
        )
        check_argument(
            item,
            expected_type=str,
            expected_choices=items,
            handle_with=IncorrectArgumentError,
            message=(
                "Argument item must be str from among "
                f"{_str_iterable(items)}"
            ),
        )
        check_if_not(
            which == "memory" and item == "number",
            IncorrectArgumentError,
            "For memory tests, you can only set repeat, not number",
        )

        return self.settings.get(func, self.defaults).get(which).get(item)

    def benchmark_function(self):
        """In-built function for benchmarking.

        It's basically an empty function that does nothing, so it's cost
        (time and memory) represent the cost of calling a function. Anything
        more, thus, results from whatever the function is doing.
        """
        pass

    def _benchmark_time(self):
        """Run timeit.repeat for the in-built benchmark function.

        Do note that unlike timeit functions, this one returns a time
        of one function call, so is normalized by "number".

        Returns:
            float: the minimal time from time.repeat per one function call
        """
        self.time_benchmark = (
            min(
                timeit.repeat(
                    lambda: self.benchmark_function,
                    number=self.defaults["time"]["number"],
                    repeat=self.defaults["time"]["repeat"],
                )
            )
            / self.defaults["time"]["number"]
        )

    def _benchmark_memory(self):
        """Run memory_profiler.memory_usage for the in-built benchmark function.

        Returns:
            float: the minimum maximum memory usage over time across all runs.
        """
        memory_results = [
            memory_usage((self.benchmark_function, (), {}))
            for _ in range(self.defaults["memory"]["repeat"])
        ]
        self.memory_benchmark = min(max(r) for r in memory_results)

    def set_defaults(self, which, number=None, repeat=None, Number=None, Repeat=None):
        """Change the default settings.

        Beware! This does not change particular settings for a particular test,
        but the defaults. So this will affect also how the benchmark tests
        will be run.

        However, this will not change any settings for functions that were already
        set using self.set, or which were already used in perftester.

        Args:
            which (str): either "time" or "memory"
            number (int, optional): passed to timeit.repeat as number.
                Defaults to None.
            repeat (int, optional): passed to timeit.repeat as repeat.
                Defaults to None.
        """
        if number is None and Number is not None:
            number = Number
        if repeat is None and Repeat is not None:
            repeat = Repeat

        self._check_args(lambda: 0, which, number, repeat)

        if number is not None:
            self.defaults[which]["number"] = number
        if repeat is not None:
            self.defaults[which]["repeat"] = repeat

    def set(self, func, which, number=None, repeat=None, Number=None, Repeat=None):
        """Set a particular argument.

        Args:
            func (Callable): a callable for which we want to change
                the setting
            which (str): either "time" or "memory"
            number (int, optional): passed to timeit.repeat as number.
                Defaults to None.
            repeat (int, optional): for time tests, passed to timeit.repeat
                as repeat; for memory tests, the number of runs of the test.
                Defaults to None.
        """
        if number is None and Number is not None:
            number = Number
        if repeat is None and Repeat is not None:
            repeat = Repeat

        self._check_args(func, which, number, repeat)

        if func not in self.settings.keys():
            self.settings[func] = {}
            self.settings[func]["memory"] = dict(
                repeat=self.defaults["memory"]["repeat"]
            )
            self.settings[func]["time"] = dict(
                number=self.defaults["time"]["number"],
                repeat=self.defaults["time"]["repeat"],
            )

        if number is not None:
            self.settings[func][which]["number"] = number
        # Below, which must be "memory"; this is checked by ._check_args() above,
        # so we do not have to worry about that here.
        if repeat is not None:
            self.settings[func][which]["repeat"] = repeat

    def _check_args(self, func, which, number, repeat):
        """Check instances of arguments func, which, number and repeat."""
        check_instance(
            func,
            Callable,
            IncorrectArgumentError,
            message=(
                "Argument func must be an int (or None), not "
                f"{type(func).__name__}"
            ),
        )

        whiches = ("time", "memory")
        if which is not None:
            check_argument(
                which,
                expected_type=str,
                expected_choices=whiches,
                handle_with=IncorrectArgumentError,
                message=(
                    "Argument which must be str from among "
                    f"{_str_iterable(whiches)}"
                ),
            )
            check_if_not(
                which == "memory" and number is not None,
                IncorrectArgumentError,
                "For memory tests, you can only set repeat, not number.",
            )

        if number is not None:
            if int(number) == number:
                number = int(number)
        if repeat is not None:
            if int(repeat) == repeat:
                repeat = int(repeat)
        check_instance(
            number,
            (int, None),
            IncorrectArgumentError,
            message=(
                "Argument number must be an int (or None), not "
                f"{type(number).__name__}"
            ),
        )
        check_instance(
            repeat,
            (int, None),
            IncorrectArgumentError,
            message=(
                "Argument repeat must be an int (or None), not "
                f"{type(repeat).__name__}"
            ),
        )


# Create a single instance of Config; this is done when the perftester module is imported.
config = Config()


def _str_iterable(an_iterable):
    """Format an_iterable as a string with items seperated with commas.

    Args:
        an_iterable (iterable): an iterable to be printed as
            a comma-seperated list

    Returns:
        str: a string with a comma-separated list of elements of an_iterable

    >>> _str_iterable({'x', 'z', 'a'})
    'a, x, z'
    >>> _str_iterable(set({'a': 20, 'g': 300, 'eeee': 300}.keys()))
    'a, eeee, g'
    """
    return ", ".join((str(i) for i in sorted(list(an_iterable))))


def _repeat(func, *args, Number=None, Repeat=None, **kwargs):
    """Run timeit.repeat for func.

    This is a simple wrapper for timeit.repeat which uses config.
    Note that the results are normalized by the number of runs,
    so the actual result is the mean execution time of a single
    run of func.

    Args:
        func (Callable): a callable to be tested

    Returns:
        list: the results of timeit.repeat
    """
    number = Number or config.get_setting(func, "time", "number")
    repeat_results = timeit.repeat(
        lambda: func(*args, **kwargs),
        number=number,
        repeat=Repeat or config.get_setting(func, "time", "repeat"),
    )
    return [r / number for r in repeat_results]


def time_test(func,
              *args,
              raw_limit=None,
              relative_limit=None,
              Number=None,
              Repeat=None,
              **kwargs
):
    """Run time performance test for func.

    You need to provide either raw_limit or relative_limit, or both.

    Note that if the tested function throws an error, so will time_test(),
    which will raise the FunctionError exception.

    When you use Number and Repeat, they have a higher priority than
    the corresponding settings from config.settings, and so they will be used.
    They are used in this single call only, and so it does not overwrite
    the config settings.

    Param:
        func (Callable): a function (or any callable) to be tested; passed
            to timeit.repeat
        raw_limit (float): raw limit for the test. It is the maximum time
            that a single run of the function can take; if it's exceeded,
            the test does not pass.
        relative_limit (float): relative limit for the test. It is used for
            relative testing (against a standard function defined in the
            config)
        args, kwargs: arguments passed to func

    Return:
        Nothing when the test passes; TimeTestError is raised otherwise
        (it then provides the limit and the measured time)

    >>> def f(n): return list(range(n))
    >>> first_run = time_benchmark(f, n=10)

    Note that this should have added the function to the config:
    >>> config.settings.keys() #doctest: +ELLIPSIS
    dict_keys([<function ...
    >>> assert_instance(first_run['raw_times'], list)
    >>> assert_instance(first_run['raw_times'][0], float)

    First, let's conduct only raw tests, comparing raw execution times (in sec).
    >>> first_run['min']  < 1e-05
    True
    >>> time_test(f, raw_limit=1e-04, n=1000)
    >>> time_test(f, raw_limit=1e-10, n=1000) #doctest: +ELLIPSIS
    Traceback (most recent call last):
       ...
    perftester.TimeTestError: Time test not passed for function f:
    raw_limit = 1e-10
    minimum run time = ...

    You can see what is the time for the benchmark function from config using
    config.benchmark_time. We cannot do it here, in the doctest, as it depends
    on the machine. However, you can use it to learn what to expect on your
    machine. You use it so in the following way (which here means that the
    tested function should not be slower than the benchmark time from config -
    that's why 1 in limits):
    >>> time_test(f, relative_limit=1, n=10) #doctest: +ELLIPSIS
    Traceback (most recent call last):
       ...
    perftester.TimeTestError: Time test not passed for function f:
    relative_limit = 1
    minimum time ratio = ...

    In our case, the test does not passes (correctly). It will surely pass if we use
    a factor of 10:
    >>> time_test(f, relative_limit=10, n=10)

    This approach based on relative becnhmarks is usuallyt what you want to
    use, since it does not use with direct time (which depends on the machine).
    Instead, we compare our the execution time of the tested function with
    the execution time of the in-built standard function (defined in the
    config), run in the same machine in which the test is run.

    Let's see if the tested function is ten times quicker than the bechmark (it is not):
    >>> time_test(f, relative_limit=.1, n=10)  #doctest: +ELLIPSIS
    Traceback (most recent call last):
       ...
    perftester.TimeTestError: Time test not passed for function f:
    relative_limit = 0.1
    minimum time ratio = ...

    """
    check_if_not(
        raw_limit is None and relative_limit is None,
        LackingLimitsError,
        message="You must provide raw_limit, relative_limit or both",
    )
    _add_func_to_config(func)
    

    results = time_benchmark(
        func, *args, Number=Number, Repeat=Repeat, **kwargs
    )

    config.cut_traceback()
    # Test raw_limit
    if raw_limit is not None:
        check_if(
            results["min"] <= raw_limit,
            handle_with=TimeTestError,
            message=(
                f"Time test not passed for function {func.__name__}:\n"
                f"raw_limit = {raw_limit}\n"
                f"minimum run time = {rounder.signif(results['min'], config.digits_for_printing)}"
            ),
        )

    # Test the relative time (against the benchmark function)
    if relative_limit is not None:
        ratio_time = results["min"] / config.time_benchmark
        check_if(
            ratio_time <= relative_limit,
            handle_with=TimeTestError,
            message=(
                f"Time test not passed for function {func.__name__}:\n"
                f"relative_limit = {relative_limit}\n"
                f"minimum time ratio = {rounder.signif(ratio_time, config.digits_for_printing)}"
            ),
        )
    config.full_traceback()


def memory_usage_test(
    func, *args, raw_limit=None, relative_limit=None, Repeat=None, **kwargs,
):
    """Test memory usage of a function.

    You can run a test based on a raw limit for the time of executing
    of the function (this is a time per single run); then use raw_limit to
    provide the limit and relative_limit=None. You can use relative_limit to set up
    a limit for a relative test (against a function defined in the config);
    then use raw_limit=None. But you can also use both at the same time, and
    the tests passes only when both the raw and the relative tests pass.

    When you use Repeat, it has a higher priority than the corresponding
    setting from config.settings, and it will be used. This is used in this
    single call only, and so it does not overwrite the config settings.

    Args:
        func (Callable): the tested function.
        raw_limit (float): raw limit for the test. It is the maximum memory
            that the function can use; if it's exceeded, the test does not pass.
        relative_limit (float): relative limit for the test. It is used for
            relative testing (against a standard function defined in the
            config)
        memory_limits (limits): an instance of namedtupe limits, with
            raw_limit and relative_limit attributes. The former is the maximum
            memory that function can use; if it's exceeded, the test does not
            pass. The latter is used for relative benchmarking (against
            a standard function defined in the config); if memory_limits is
            None, the function will provide the memory usage times, which can
            be used to define the limits to be used in testing
        Repeat (int): a repeat setting, overwriting (for this function call)
            the setting in config.settings[func]["memory"]["repeat"]
        *args, **kwargs: arguments passed to the function
     Returns:
        when memory_limit is None, a dict:
            'raw_results': all results (through time), as returned from several
                runs of memory_profile.memory_usage
            'mean_result_per_run': a list of mean results per run
            'max_result_per_run': the maximum result per run (which is perhaps
                the most interesting to us)
            'mean': the mean of the means
            'max': the overall max (so, the maximum usage of memory across all
                runs)
        when memory_limits is not None:
            Nothing when the test passes; perftesterError is raised otherwise
            (it then provides the limit and the measured memory usage)
    >>> def sum1(n): return sum([i**2 for i in range(n)])
    >>> first_run = memory_usage_benchmark(sum1, n=100_000)
    >>> first_run.keys()
    dict_keys(['raw_results', 'relative_results', 'mean_result_per_run', 'max_result_per_run', 'max_result_per_run_relative', 'mean', 'max', 'max_relative'])
    >>> type(first_run['raw_results'])
    <class 'list'>
    >>> first_run['mean'] < first_run['max']
    True
    >>> memory_usage_test(sum1, raw_limit=first_run['max']*2, n=100_000)
    """
    check_instance(
        func,
        Callable,
        IncorrectArgumentError,
        "Argument func must be a callable.",
    )
    check_if_not(
        raw_limit is None and relative_limit is None,
        LackingLimitsError,
        message="You must provide raw_limit, relative_limit or both",
    )
    results = memory_usage_benchmark(func, *args, Repeat=Repeat, **kwargs)

    config.cut_traceback()
    if raw_limit is not None:
        check_if(
            results["max"] <= raw_limit,
            handle_with=MemoryTestError,
            message=(
                f"Memory test not passed for function {func.__name__}:\n"
                f"memory_limit = {raw_limit}\n"
                f"maximum memory usage = {rounder.signif(results['max'], config.digits_for_printing)}"
            ),
        )
    if relative_limit is not None:
        relatve_got_memory = results["max"] / config.memory_benchmark
        check_if(
            relatve_got_memory <= relative_limit,
            handle_with=MemoryTestError,
            message=(
                f"Memory test not passed for function {func.__name__}:\n"
                f"relative memory limit = {relative_limit}\n"
                f"maximum obtained relative memory usage = "
                f"{rounder.signif(relatve_got_memory, config.digits_for_printing)}"
            ),
        )
    config.full_traceback()


def memory_usage_benchmark(func, *args, Repeat=None, **kwargs):
    """Run memory benchmarks for a callable.

    It offers an easy way to run benchmarks before defining limits in
    memory_usage_test(). The benchmarks use the settings from config for func,
    or defaults if this function has no settings yet. Note that after
    running the benchmark function for a function, this function will
    be added to config.settings.

    When you use Repeat, it has a higher priority than the corresponding
    setting from config.settings, and it will be used. This is used in this
    single call only, and so it does not overwrite the config settings.

    The function returns a dict that you can pretty-print using function pp().

    Args:
        func (Callable): a function (or a callable) to run benchmarks for
        Repeat (int): a repeat setting, overwriting (for this function call)
            the setting in config.settings[func]["memory"]["repeat"]
        *args, **kwargs: arguments passed to func

    Returns:
        dict: a dictionary with results returned by time_test() and
            memory_usage_test()

    >>> def f(x): return x
    >>> f_bench = memory_usage_benchmark(f, x=10)
    >>> f_bench.keys()
    dict_keys(['raw_results', 'relative_results', 'mean_result_per_run', 'max_result_per_run', 'max_result_per_run_relative', 'mean', 'max', 'max_relative'])
    """
    check_instance(func, Callable, message="Argument func must be a callable.")
    _add_func_to_config(func)

    n = Repeat or config.settings[func]["memory"]["repeat"]

    try:
        memory_results = [memory_usage((func, args, kwargs)) for i in range(n)]
    except Exception as e:
        raise FunctionError(
            f"The tested function raised {type(e).__name__}: {str(e)}"
        )

    memory_results_mean = [mean(this_result) for this_result in memory_results]
    memory_results_max = [max(this_result) for this_result in memory_results]
    overall_mean = mean(memory_results_mean)
    # We take the min of the max values
    overall_max = min(memory_results_max)

    relative_results = copy.deepcopy(memory_results)
    for i, result in enumerate(relative_results):
        for j, r in enumerate(result):
            relative_results[i][j] = r / config.memory_benchmark
    return {
        "raw_results": memory_results,
        "relative_results": relative_results,
        "mean_result_per_run": memory_results_mean,
        "max_result_per_run": memory_results_max,
        "max_result_per_run_relative": [
            r / config.memory_benchmark for r in memory_results_max
        ],
        "mean": overall_mean,
        "max": overall_max,
        "max_relative": overall_max / config.memory_benchmark,
    }


def time_benchmark(func, *args, Number=None, Repeat=None, **kwargs):
    """Run time benchmarks for a callable.

    It offers an easy way to run benchmarks before defining limits in
    time_test(). The benchmarks use the settings from config for func,
    or defaults if this function has no settings yet. Note that after
    running the benchmark function for a function, this function will
    be added to config.settings.

    The function returns a dict that you can pretty-print using function pp().

    Args:
        func (Callable): a function (or a callable) to run benchmarks for
        Number (int): a number setting, overwriting (for this function call)
            the setting in config.settings[func]["memory"]["number"]
        Repeat (int): a repeat setting, overwriting (for this function call)
            the setting in config.settings[func]["time"]["repeat"]
        *args, **kwargs: arguments passed to func

    Returns:
        dict:
            'raw_times': all the received times (from each repeat)
            'raw_times_relative': all the received times (from each repeat),
                divided by config.benchmark_time
            'min': the minimum time (from the repeats)
            'mean': mean times
            'max': the maximum time
            'max_relative': the maximum time, divided by config.benchmark_time

    >>> def f(x): return x
    >>> f_bench = time_benchmark(f, x=10)
    >>> f_bench.keys()
    dict_keys(['min', 'min_relative', 'raw_times', 'raw_times_relative', 'mean', 'max'])

    Of course, we do have to remember that the execution time can depend on
    the function's arguments:
    >>> def f(n): return list(range(n))
    >>> time_benchmark(f, n=1000)["min"] > time_benchmark(f, n=1)["min"]
    True

    """
    check_instance(func, Callable, message="Argument func must be a callable.")
    _add_func_to_config(func)

    try:
        results = _repeat(func, *args, Number=Number, Repeat=Repeat, **kwargs)
    except Exception as e:
        raise FunctionError(
            f"The tested function raised {type(e).__name__}: {str(e)}"
        )
    
    min_result = min(results)

    # Reload built-in benchmarks
    config.reload_time()

    return {
        "min": min_result,
        "min_relative": min_result / config.time_benchmark,
        "raw_times": results,
        "raw_times_relative": [
            result / config.time_benchmark for result in results
        ],
        "mean": mean(results),
        "max": max(results),
    }


def pp(*args):
    """Pretty print of args.

    If any numbers are among the data, they are rounded to the number of
    significant digits as given by config.digits_for_printing (the default
    is 4). If more than one argument is provided, they are printed with a
    line break used as a separator.

    >>> pp(.122242)
    0.1222
    >>> pp(dict(a=.12121212, b=23.234234234), ["system failure", 345345.345])
    {'a': 0.1212, 'b': 23.23}
    ['system failure', 345300.0]
    """
    for arg in args:
        pprint(
            rounder.signif_object(
                arg, digits=config.digits_for_printing, use_copy=True
            )
        )


def _add_func_to_config(func):
    if func not in config.settings.keys():
        config.settings[func] = {}
        config.settings[func]["time"] = dict(
            number=config.defaults["time"]["number"],
            repeat=config.defaults["time"]["repeat"],
        )
        config.settings[func]["memory"] = dict(
            repeat=config.defaults["memory"]["repeat"],
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod()
